/*
   LCD routines.
   Assembly wrappers for access to the RCX ROM based routines.
   Last updated 2/25/2019
 */

#include "H8.h"


#ifndef   LCD_H_DEFINED
#define   LCD_H_DEFINED

//! helper macros
//
#define BYTE_OF(a,b) a
#define BIT_OF(a,b)	b

// LCD segment control byte and bit locations
// 0xNNNN,0xM => Mth bit (value 1<<M) of byte 0xNNNN
// overall memory range: 0xef43-0xef4b (9 bytes)
//
//! set a segment directly in the LCD buffer
/*! \param a the segment to set
 */
#define dlcd_show(a) bit_set(BYTE_OF(a),BIT_OF(a))

//! clear a segment directly in the LCD buffer
/*! \param a the segment to clear
 */
#define dlcd_hide(a) bit_clear(BYTE_OF(a),BIT_OF(a))

//! store the carry flag to a segment directly in the LCD buffer
/*! \param a the segment to store to

   this is highly useful in combination with bit_load(mask,bit)
 */
#define dlcd_store(a) bit_store(BYTE_OF(a),BIT_OF(a))

// the man
//
#define LCD_ARMS	0xef43,0x0
#define LCD_BODY	0xef43,0x1
#define LCD_1LEG	0xef43,0x2
#define LCD_2LEGS	0xef43,0x3

// the digits
// numbered right to left
// dots are associated with the digit to their left
// only digit 2,3,4 have dots
// digit 5 can only be used as a minus sign
//
#define LCD_0_TOP	0xef44,0x4
#define LCD_0_MID	0xef44,0x5
#define LCD_0_BOT	0xef44,0x7
#define LCD_0_TOPR	0xef45,0x5
#define LCD_0_BOTR	0xef45,0x7
#define LCD_0_TOPL	0xef47,0x5
#define LCD_0_BOTL	0xef47,0x7

#define LCD_1_TOP	0xef43,0x4
#define LCD_1_MID	0xef43,0x5
#define LCD_1_BOT	0xef43,0x7
#define LCD_1_TOPR	0xef47,0x1
#define LCD_1_BOTR	0xef47,0x3
#define LCD_1_TOPL	0xef48,0x5
#define LCD_1_BOTL	0xef48,0x7

#define LCD_2_TOP	0xef44,0x0
#define LCD_2_MID	0xef44,0x1
#define LCD_2_BOT	0xef44,0x3
#define LCD_2_TOPR	0xef48,0x1
#define LCD_2_BOTR	0xef48,0x3
#define LCD_2_TOPL	0xef49,0x5
#define LCD_2_BOTL	0xef49,0x7
#define LCD_2_DOT	0xef48,0x6

#define LCD_3_TOP	0xef46,0x0
#define LCD_3_MID	0xef46,0x1
#define LCD_3_BOT	0xef46,0x3
#define LCD_3_TOPR	0xef4b,0x1
#define LCD_3_BOTR	0xef4b,0x3
#define LCD_3_TOPL	0xef4b,0x5
#define LCD_3_BOTL	0xef4b,0x7
#define LCD_3_DOT	0xef49,0x6

#define LCD_4_TOP	0xef46,0x4
#define LCD_4_MID	0xef46,0x5
#define LCD_4_BOT	0xef46,0x7
#define LCD_4_TOPR	0xef4a,0x1
#define LCD_4_BOTR	0xef4a,0x3
#define LCD_4_TOPL	0xef4a,0x5
#define LCD_4_BOTL	0xef4a,0x7
#define LCD_4_DOT	0xef4b,0x6

#define LCD_5_MID	0xef4a,0x6

// the motor displays
#define LCD_A_SELECT	0xef4a,0x2
#define LCD_A_LEFT	0xef46,0x6
#define LCD_A_RIGHT	0xef46,0x2

#define LCD_B_SELECT	0xef43,0x6
#define LCD_B_LEFT	0xef44,0x2
#define LCD_B_RIGHT	0xef47,0x2

#define LCD_C_SELECT	0xef44,0x6
#define LCD_C_LEFT	0xef47,0x6
#define LCD_C_RIGHT	0xef45,0x6

// the sensor displays
#define LCD_S1_SELECT	0xef49,0x0
#define LCD_S1_ACTIVE	0xef49,0x1

#define LCD_S2_SELECT	0xef48,0x4
#define LCD_S2_ACTIVE	0xef48,0x0

#define LCD_S3_SELECT	0xef47,0x4
#define LCD_S3_ACTIVE	0xef45,0x4

// the quartered circle
// quarter nubering is counterclockwise
// starts at top right
#define LCD_CIRCLE_0	0xef45,0x0
#define LCD_CIRCLE_1	0xef45,0x1
#define LCD_CIRCLE_2	0xef45,0x3
#define LCD_CIRCLE_3	0xef45,0x2

// the dotted line
// dots numbered right to left
#define LCD_DOT_0	0xef49,0x4
#define LCD_DOT_1	0xef4b,0x0
#define LCD_DOT_2	0xef4b,0x4
#define LCD_DOT_3	0xef4a,0x0
#define LCD_DOT_4	0xef4a,0x4

// the IR display
#define LCD_IR_LOWER	0xef49,0x2
#define LCD_IR_UPPER	0xef49,0x3

// the crossed out battery
#define LCD_BATTERY_X	0xef47,0x0

// two empty bits
#define LCD_EMPTY_1	0xef48,0x2
#define LCD_EMPTY_2	0xef4b,0x2

const char hex_display_codes[] =
{
    0x7e,			// 0
    0x42,			// 1
    0x37,			// 2
    0x67,			// 3
    0x4b,			// 4
    0x6d,			// 5
    0x7d,			// 6
    0x46,			// 7
    0x7f,			// 8
    0x6f,			// 9
    0x5f,			// A
    0x79,			// b
    0x31,			// c
    0x73,			// d
    0x3d,			// E
    0x1d,			// F
};

typedef enum
{
    man_stand = 0x3006,
    man_run   = 0x3007,
    s1_select = 0x3008,
    s1_active = 0x3009,
    s2_select = 0x300A,
    s2_active = 0x300B,
    s3_select = 0x300C,
    s3_active = 0x300D,
    a_select  = 0x300E,
    a_left    = 0x300F,
    a_right   = 0x3010,
    b_select  = 0x3011,
    b_left    = 0x3012,
    b_right   = 0x3013,
    c_select  = 0x3014,
    c_left    = 0x3015,
    c_right   = 0x3016,
    unknown_1 = 0x3017,		//!< seemingly without effect. cycle reset?
    circle    = 0x3018,		//!< 0..3 quarters: add one. 4 quarters: reset
    dot       = 0x3019,		//!< 0..4 dots: add a dot. 5 dots: reset
    // dots are added on the right.
    dot_inv   = 0x301A,		//!< 0 dots: show 5. 1..4 dots: subtract one
    // dots are removed on the left
    battery_x = 0x301B,
    ir_half   = 0x301C, 	    	//! the IR display values are mutually exclusive.
    ir_full   = 0x301D,   	//! the IR display values are mutually exclusive.
    everything= 0x3020

} lcd_segment;

//! LCD number display styles
/*! note: signed and unsigned are taken by the C programming language
 */
typedef enum
{
    digit  = 0x3017,		//!< single digit on the right
    // works only with comma_style digit
    sign   = 0x3001,		//!< signed, no leading zeros
    unsign = 0x301F		//!< unsigned, 0 displayed as 0000

} lcd_number_style;

//! LCD comma display styles
/*
 */
typedef enum
{
    digit_comma = 0x0000,		//!< single digit on the right
    // works only with number_style digit
    e0  = 0x3002,			//!< whole
    e_1 = 0x3003,	  	      	//!< 10ths
    e_2 = 0x3004,		      	//!< 100ths
    e_3 = 0x3005,		      	//!< 1000ths, problematic with negatives
    e_4 = 0x3006
} lcd_comma_style;		// lcd display comma style

void lcd_refresh (void)
{
    asm ("jsr 0x27c8      ; call refresh \n\t");
}

void lcd_clear (void)
{
    asm (
        "jsr 0x27ac      ; call clear \n\t"
        "jsr 0x27c8      ; call refresh \n\t"
    );
}

void lcd_show_icon (int icon)
{
    asm (
        "mov.w r0,r6 \n\t"
        "jsr 0x1b62      ; call show_icon \n\t"
        "jsr 0x27c8      ; call refresh \n\t"
    );
}

void lcd_hide_icon (int icon)
{
    asm (
        "mov.w r0,r6 \n\t"
        "jsr 0x1e4a      ; call show_icon \n\t"
        "jsr 0x27c8      ; call refresh \n\t"
    );
}

/* Shows integer value as four decimal digits in the LCD left of
 * the LEGO man. This means that the range is [-9999, 9999].
 * Negative values are show with a leading minus sign.
 */

//! display an integer in decimal
#define lcd_int(i)	lcd_number(i,sign,e0)
//#define lcd_int(i)	lcd_number(i,0x3001,0x3002)

//! display an unsigned value in decimal
#define lcd_unsigned(u)	lcd_number(u,unsign,e0)

//! display a clock.
/*! passing an argument of 1015 will display 10.15
 */
#define lcd_clock(t)	lcd_number(t,unsign,e_2)

//! display a single digit right of the man symbol
#define lcd_digit(d)	lcd_number(d,digit,digit_comma)

void lcd_number(int i, lcd_number_style n, lcd_comma_style c)
{
    asm volatile (
        "push r2	; lcd_comma_style \n\t"
        "push r0	; i \n\t"
        "mov.w r1,r6	; lcd_number_style -> r6 \n\t"
        "jsr 0x1ff2      ; call show_number \n\t"
        "adds #2,r7      ; adjust stack \n\t"
        "adds #2,r7 \n\t"
    );
}

// gcc is stupid
// doesn't re-use constant values in registers.
// re-ordered stores to help him.
void cputc_native(char mask, int pos)
{
    switch (pos)
    {
    case 0:
        //! display native mode segment mask at display position 0
        /*! \param mask the mask to display
        */
        bit_load(mask, 0x2);
        dlcd_store(LCD_0_TOP);
        bit_load(mask, 0x0);
        dlcd_store(LCD_0_MID);
        bit_load(mask, 0x5);
        dlcd_store(LCD_0_BOT);
        bit_load(mask, 0x1);
        dlcd_store(LCD_0_TOPR);
        bit_load(mask, 0x6);
        dlcd_store(LCD_0_BOTR);
        bit_load(mask, 0x3);
        dlcd_store(LCD_0_TOPL);
        bit_load(mask, 0x4);
        dlcd_store(LCD_0_BOTL);
        break;
    case 1:
        //! display native mode segment mask at display position 1
        /*! \param mask the mask to display
        */
        bit_load(mask, 0x2);
        dlcd_store(LCD_1_TOP);
        bit_load(mask, 0x0);
        dlcd_store(LCD_1_MID);
        bit_load(mask, 0x5);
        dlcd_store(LCD_1_BOT);
        bit_load(mask, 0x1);
        dlcd_store(LCD_1_TOPR);
        bit_load(mask, 0x6);
        dlcd_store(LCD_1_BOTR);
        bit_load(mask, 0x3);
        dlcd_store(LCD_1_TOPL);
        bit_load(mask, 0x4);
        dlcd_store(LCD_1_BOTL);
        break;
    case 2:
        //! display native mode segment mask at display position 2
        /*! \param mask the mask to display
         */
        bit_load(mask, 0x2);
        dlcd_store(LCD_2_TOP);
        bit_load(mask, 0x0);
        dlcd_store(LCD_2_MID);
        bit_load(mask, 0x5);
        dlcd_store(LCD_2_BOT);
        dlcd_hide(LCD_2_DOT);
        bit_load(mask, 0x1);
        dlcd_store(LCD_2_TOPR);
        bit_load(mask, 0x6);
        dlcd_store(LCD_2_BOTR);
        bit_load(mask, 0x3);
        dlcd_store(LCD_2_TOPL);
        bit_load(mask, 0x4);
        dlcd_store(LCD_2_BOTL);
        break;
    case 3:
        //! display native mode segment mask at display position 3
        /*! \param mask the mask to display
         */
        dlcd_hide(LCD_3_DOT);
        bit_load(mask, 0x2);
        dlcd_store(LCD_3_TOP);
        bit_load(mask, 0x0);
        dlcd_store(LCD_3_MID);
        bit_load(mask, 0x5);
        dlcd_store(LCD_3_BOT);
        bit_load(mask, 0x1);
        dlcd_store(LCD_3_TOPR);
        bit_load(mask, 0x6);
        dlcd_store(LCD_3_BOTR);
        bit_load(mask, 0x3);
        dlcd_store(LCD_3_TOPL);
        bit_load(mask, 0x4);
        dlcd_store(LCD_3_BOTL);
        break;
    case 4:
        //! display native mode segment mask at display position 4
        /*! \param mask the mask to display
         */
        dlcd_hide(LCD_4_DOT);
        bit_load(mask, 0x2);
        dlcd_store(LCD_4_TOP);
        bit_load(mask, 0x0);
        dlcd_store(LCD_4_MID);
        bit_load(mask, 0x5);
        dlcd_store(LCD_4_BOT);
        bit_load(mask, 0x1);
        dlcd_store(LCD_4_TOPR);
        bit_load(mask, 0x6);
        dlcd_store(LCD_4_BOTR);
        bit_load(mask, 0x3);
        dlcd_store(LCD_4_TOPL);
        bit_load(mask, 0x4);
        dlcd_store(LCD_4_BOTL);
        break;
    case 5:
        //! display native mode segment mask at display position 5
        /*! \param mask the mask to display. only the middle segment is
           present on the display.
         */
        bit_load(mask, 0x0);
        dlcd_store(LCD_5_MID);
    }
}


//! Write HEX digit to specified position of LCD
/*! (this is essentially a dispatcher for cputc_hex_[0-5] functions)
 *  \param c the HEX digit to be displayed
 *  \param pos the location at which to display the HEX digit
 *  \return Nothing
 */
void cputc_hex(char c, int pos)
{
    cputc_native(hex_display_codes[(c) & 0x7f], pos);
}

//! display a hexword in the four leftmost positions.
/*! \param word the hexword

   position 0 is unaffected by this call.
 */
void cputw(unsigned word)
{
	int i;
    //Disable;

    cputc_native(0, 5);

    for (i = 1; i <= 4; i++)
    {
        cputc_hex(word & 0x0f, i);
        word >>= 4;
    }

    lcd_refresh();
    //Enable;
}


#endif /* LCD_H_DEFINED */
